from Orbaplaw.NaturalBondOrbitalMethods import NaturalAtomicOrbital
from Orbaplaw.NaturalBondOrbitalMethods import NaturalBondOrbital
from Orbaplaw.NaturalBondOrbitalMethods import PrincipalInteractingOrbital

generateNaturalAtomicOrbital=NaturalAtomicOrbital.generateNaturalAtomicOrbital
NaturalAtomicOrbital=NaturalAtomicOrbital.NaturalAtomicOrbital
generateNaturalBondOrbital=NaturalBondOrbital.generateNaturalBondOrbital
NaturalBondOrbital=NaturalBondOrbital.NaturalBondOrbital
generatePrincipalInteractingOrbital=PrincipalInteractingOrbital.generatePrincipalInteractingOrbital
PrincipalInteractingOrbital=PrincipalInteractingOrbital.PrincipalInteractingOrbital
